use crate::assets::{draw_string, Assets};
use crate::config::{SCREEN_HEIGHT, SCREEN_WIDTH};
use crate::menu::GameMenu;
use crate::particle::{
    ConfigPopUpBasic, ConfigRingBasic, ParticleControllerPopUp, ParticleControllerRing,
    ParticleType, ParticleWorld,
};
use crate::sky::{Cloud, Sky, Sun};
use macroquad::audio::play_sound_once;
use macroquad::color;
use macroquad::color::Color;
use macroquad::input::{is_key_down, KeyCode};
use macroquad::math::{Rect, Vec2};
use macroquad::prelude::{draw_texture, draw_texture_ex, is_key_pressed, vec2, DrawTextureParams};
use macroquad::rand::RandomRange;
use macroquad::shapes::draw_hexagon;
use macroquad::time::{get_frame_time, get_time};
use std::collections::LinkedList;

pub enum GameState<'a> {
    Game(GameWorld<'a>),
    Menu(GameMenu),
}

trait BoundingBox {
    fn get_bounds(&self) -> Option<Rect>;

    fn can_enter_from_dir(&self, angle: f32) -> bool;

    fn replace_with_closed(&mut self);
}

enum WindowAttach {
    PinkCurtain,
    CyanCurtain,
    MiniCat,
    Plants,
    YellowCurtain,
    ActiveBucket,
    Bucket,
}

impl WindowAttach {
    fn generate() -> Self {
        match u32::gen_range(0, 200) {
            ..=30 => WindowAttach::PinkCurtain,
            31..=60 => WindowAttach::CyanCurtain,
            61..=90 => WindowAttach::YellowCurtain,
            91..=120 => WindowAttach::Plants,
            121..=200 => WindowAttach::ActiveBucket,
            201..=220 => WindowAttach::MiniCat,
            _ => WindowAttach::MiniCat,
        }
    }

    fn attach_offset(&self) -> Vec2 {
        match *self {
            WindowAttach::Bucket => vec2(1.0, 0.0),
            WindowAttach::ActiveBucket => vec2(2.0, 1.0),
            _ => vec2(0.0, 0.0),
        }
    }

    fn sprite_rect(&self) -> Rect {
        match *self {
            WindowAttach::PinkCurtain => Rect::new(1.0, 65.0, 13.0, 11.0),
            WindowAttach::CyanCurtain => Rect::new(15.0, 65.0, 13.0, 11.0),
            WindowAttach::MiniCat => Rect::new(29.0, 65.0, 13.0, 11.0),
            WindowAttach::Plants => Rect::new(43.0, 65.0, 13.0, 11.0),
            WindowAttach::YellowCurtain => Rect::new(57.0, 65.0, 13.0, 11.0),
            WindowAttach::Bucket => Rect::new(2.0, 79.0, 15.0, 12.0),
            WindowAttach::ActiveBucket => Rect::new(19.0, 78.0, 17.0, 14.0),
        }
    }

    fn render(&self, offset: Vec2, assets: &Assets) {
        let pos = offset - self.attach_offset();

        let sprite_rect = self.sprite_rect();
        draw_texture_ex(
            assets.house_texture,
            pos.x,
            pos.y,
            color::WHITE,
            DrawTextureParams {
                source: Some(sprite_rect),
                ..Default::default()
            },
        );

        let t = get_time();
        if let WindowAttach::ActiveBucket = self {
            let step = (t * 2.0 % 4.0).floor() as usize;
            let steps = [0.0, 1.0, 2.0, 3.0];

            draw_texture_ex(
                assets.projectiles_texture,
                offset.x + sprite_rect.w / 2.0 - 9.0,
                offset.y - 19.0 + (t * 2.0).sin() as f32,
                color::WHITE,
                DrawTextureParams {
                    source: Some(Rect::new(steps[step] as f32 * 16.0, 48.0, 16.0, 16.0)),
                    ..Default::default()
                },
            );
        }
    }
}

impl BoundingBox for WindowAttach {
    fn get_bounds(&self) -> Option<Rect> {
        match *self {
            WindowAttach::ActiveBucket => Some(Rect::new(-3.0, -2.0, 19.0, 2.0)),
            _ => None,
        }
    }

    fn can_enter_from_dir(&self, angle: f32) -> bool {
        (0.0..std::f32::consts::PI).contains(&angle)
    }

    fn replace_with_closed(&mut self) {
        *self = WindowAttach::Bucket;
    }
}

enum DoorAttach {
    Door,
    ActiveDoor,
    ActivePortalDoor,
}

impl DoorAttach {
    fn generate() -> Self {
        match u32::gen_range(0, 100) {
            ..=50 => DoorAttach::Door,
            51..=95 => DoorAttach::ActiveDoor,
            96..=100 => DoorAttach::ActivePortalDoor,
            _ => DoorAttach::ActiveDoor,
        }
    }

    fn attach_offset(&self) -> Vec2 {
        match *self {
            DoorAttach::ActiveDoor | DoorAttach::ActivePortalDoor => vec2(1.0, 1.0),
            _ => vec2(0.0, 0.0),
        }
    }

    fn sprite_rect(&self) -> Rect {
        match *self {
            DoorAttach::Door => Rect::new(3.0, 94.0, 13.0, 16.0),
            DoorAttach::ActiveDoor => Rect::new(20.0, 93.0, 15.0, 18.0),
            DoorAttach::ActivePortalDoor => Rect::new(39.0, 93.0, 15.0, 18.0),
        }
    }

    fn render(&self, offset: Vec2, assets: &Assets) {
        let pos = offset - self.attach_offset();

        let sprite_rect = self.sprite_rect();

        draw_texture_ex(
            assets.house_texture,
            pos.x,
            pos.y,
            color::WHITE,
            DrawTextureParams {
                source: Some(sprite_rect),
                ..Default::default()
            },
        );

        let t = get_time();

        match *self {
            DoorAttach::ActiveDoor | DoorAttach::ActivePortalDoor => {
                let step = (t * 2.0 % 4.0).floor() as usize;
                let steps = [0.0, 1.0, 2.0, 3.0];

                draw_texture_ex(
                    assets.projectiles_texture,
                    offset.x - 19.0 + (t * 2.0).sin() as f32,
                    offset.y + sprite_rect.h / 2.0 - 9.0,
                    color::WHITE,
                    DrawTextureParams {
                        source: Some(Rect::new(steps[step] as f32 * 16.0, 32.0, 16.0, 16.0)),
                        ..Default::default()
                    },
                );
            }
            _ => {}
        }
    }
}

impl BoundingBox for DoorAttach {
    fn get_bounds(&self) -> Option<Rect> {
        match *self {
            DoorAttach::ActiveDoor | DoorAttach::ActivePortalDoor => {
                Some(Rect::new(-3.0, -3.0, 13.0, 15.0))
            }
            _ => None,
        }
    }

    fn can_enter_from_dir(&self, angle: f32) -> bool {
        (-std::f32::consts::FRAC_PI_2..std::f32::consts::FRAC_PI_2).contains(&angle)
    }

    fn replace_with_closed(&mut self) {
        *self = DoorAttach::Door;
    }
}

struct Attach<T: BoundingBox>(Vec2, T);

enum House {
    Round {
        window: Attach<WindowAttach>,
        door: Attach<DoorAttach>,
    },
    Wooden {
        window_left: Attach<WindowAttach>,
        window_right: Attach<WindowAttach>,
        door: Attach<DoorAttach>,
    },
}

impl House {
    fn generate() -> Self {
        match u32::gen_range(0, 100) {
            ..=25 => House::Wooden {
                door: Attach(vec2(34.0, 48.0), DoorAttach::generate()),
                window_left: Attach(vec2(17.0, 33.0), WindowAttach::generate()),
                window_right: Attach(vec2(34.0, 33.0), WindowAttach::generate()),
            },
            _ => House::Round {
                door: Attach(vec2(37.0, 48.0), DoorAttach::generate()),
                window: Attach(vec2(15.0, 34.0), WindowAttach::generate()),
            },
        }
    }

    fn attach_offset(&self) -> Vec2 {
        match *self {
            House::Round { .. } => vec2(0.0, 0.0),
            House::Wooden { .. } => vec2(0.0, 0.0),
        }
    }

    fn sprite_rect(&self) -> Rect {
        match *self {
            House::Round { .. } => Rect::new(0.0, 0.0, 64.0, 64.0),
            House::Wooden { .. } => Rect::new(64.0, 0.0, 64.0, 64.0),
        }
    }

    fn render(&self, offset: Vec2, assets: &Assets) {
        let pos = offset - self.attach_offset();

        draw_texture_ex(
            assets.house_texture,
            pos.x,
            pos.y,
            color::WHITE,
            DrawTextureParams {
                source: Some(self.sprite_rect()),
                ..Default::default()
            },
        );

        match self {
            House::Round {
                door: Attach(door_attach, door),
                window: Attach(window_attach, window),
            } => {
                window.render(pos + *window_attach, assets);
                door.render(pos + *door_attach, assets);
            }
            House::Wooden {
                door: Attach(door_attach, door),
                window_left: Attach(window_left_attach, window_left),
                window_right: Attach(window_right_attach, window_right),
            } => {
                window_left.render(pos + *window_left_attach, assets);
                window_right.render(pos + *window_right_attach, assets);
                door.render(pos + *door_attach, assets);
            }
        };
    }
}

enum GrassTop {
    GrassYellowFlower,
}

impl GrassTop {
    fn generate() -> Self {
        GrassTop::GrassYellowFlower
    }

    fn sprite_rect(&self) -> Rect {
        match *self {
            GrassTop::GrassYellowFlower => Rect::new(0.0, 0.0, 64.0, 32.0),
        }
    }

    fn render(&self, offset: Vec2, assets: &Assets) {
        draw_texture_ex(
            assets.grass_top_texture,
            offset.x,
            offset.y,
            color::WHITE,
            DrawTextureParams {
                source: Some(self.sprite_rect()),
                ..Default::default()
            },
        );
    }
}

enum Road {
    Asphalt1,
    Asphalt2,
}

impl Road {
    fn generate() -> Self {
        match u32::gen_range(0, 100) {
            ..=50 => Road::Asphalt1,
            51..=100 => Road::Asphalt2,
            _ => Road::Asphalt1,
        }
    }

    fn sprite_rect(&self) -> Rect {
        match *self {
            Road::Asphalt1 => Rect::new(0.0, 0.0, 64.0, 32.0),
            Road::Asphalt2 => Rect::new(64.0, 0.0, 64.0, 32.0),
        }
    }

    fn render(&self, offset: Vec2, assets: &Assets) {
        draw_texture_ex(
            assets.road_texture,
            offset.x,
            offset.y,
            color::WHITE,
            DrawTextureParams {
                source: Some(self.sprite_rect()),
                ..Default::default()
            },
        );
    }
}

enum GrassBottom {
    GrassRedFlower,
}

impl GrassBottom {
    fn generate() -> Self {
        GrassBottom::GrassRedFlower
    }

    fn sprite_rect(&self) -> Rect {
        match *self {
            GrassBottom::GrassRedFlower => Rect::new(0.0, 0.0, 64.0, 32.0),
        }
    }

    fn render(&self, offset: Vec2, assets: &Assets) {
        draw_texture_ex(
            assets.grass_bottom_texture,
            offset.x,
            offset.y,
            color::WHITE,
            DrawTextureParams {
                source: Some(self.sprite_rect()),
                ..Default::default()
            },
        );
    }
}

struct RoadBorder;

impl RoadBorder {
    fn sprite_rect(&self) -> Rect {
        Rect::new(112.0, 32.0, 16.0, 32.0)
    }

    fn render(&self, offset: Vec2, assets: &Assets) {
        draw_texture_ex(
            assets.road_texture,
            offset.x,
            offset.y,
            color::WHITE,
            DrawTextureParams {
                source: Some(self.sprite_rect()),
                ..Default::default()
            },
        );
    }
}

enum RoadBorderPosition {
    Top(RoadBorder),
    Bottom(RoadBorder),
    None,
}

struct Slice {
    lawn: GrassBottom,
    house: Option<House>,
    road_grass_bottom: GrassBottom,
    road: Road,
    road_border: RoadBorderPosition,
    road_grass_top: GrassTop,
}

impl Slice {
    const WIDTH: f32 = 64.0;
    const HOUSE_ATTACH_POINT: Vec2 = Vec2 { x: 0.0, y: 128.0 };

    fn generate() -> Self {
        Self {
            lawn: GrassBottom::generate(),
            house: if u32::gen_range(0, 100) > 60 {
                Some(House::generate())
            } else {
                None
            },
            road_grass_bottom: GrassBottom::generate(),
            road: Road::generate(),
            road_border: match u32::gen_range(0, 100) {
                ..=10 => RoadBorderPosition::Top(RoadBorder),
                11..=21 => RoadBorderPosition::Bottom(RoadBorder),
                _ => RoadBorderPosition::None,
            },
            road_grass_top: GrassTop::generate(),
        }
    }

    fn render(&self, offset: Vec2, assets: &Assets) {
        self.road.render(offset - vec2(0.0, 52.0), assets);
        self.road_grass_bottom
            .render(offset - vec2(0.0, 32.0), assets);
        self.road_grass_top.render(offset - vec2(0.0, 64.0), assets);
        self.lawn.render(offset - vec2(0.0, 96.0), assets);

        if let Some(ref house) = self.house {
            house.render(offset - Self::HOUSE_ATTACH_POINT, assets);
        }

        if let RoadBorderPosition::Top(border) = &self.road_border {
            border.render(offset - vec2(0.0, 68.0), assets);
        }
    }

    fn render_front(&self, offset: Vec2, assets: &Assets) {
        if let RoadBorderPosition::Bottom(border) = &self.road_border {
            border.render(offset - vec2(0.0, 52.0), assets);
        }
    }
}

struct Player {
    lives: usize,
    pos: Vec2,
    speed: f64,
    start_time: f64,
    pedal_speed: f64,
    pedal_speed_mod: f64,
}

impl Player {
    fn new() -> Self {
        Player {
            pos: vec2(40.0, SCREEN_HEIGHT - 70.0),
            speed: 1.0,
            start_time: get_time(),
            pedal_speed: 1.0,
            pedal_speed_mod: 1.0,
            lives: 5,
        }
    }

    fn shoot_velocity(&self) -> Vec2 {
        vec2(60.0 * self.pedal_speed as f32, -120.0)
    }

    fn render(&mut self, assets: &Assets) {
        let t = get_time();
        self.speed = (1.0 + t - self.start_time).log(20.0);

        let dt = get_frame_time();

        if is_key_down(KeyCode::Up) || is_key_down(KeyCode::W) {
            self.pos.y -= Slice::WIDTH / 2.0 * self.speed as f32 * dt;
            self.pos.y = self.pos.y.max(SCREEN_HEIGHT - 54.0 - 29.0);
        } else if is_key_down(KeyCode::Down) || is_key_down(KeyCode::S) {
            self.pos.y += Slice::WIDTH / 2.0 * self.speed as f32 * dt;
            self.pos.y = self.pos.y.min(SCREEN_HEIGHT - 20.0 - 29.0);
        }

        let mut pedal_speed_target = 1.0;

        if is_key_down(KeyCode::Left) || is_key_down(KeyCode::A) {
            self.pos.x -= Slice::WIDTH / 2.0 * self.speed as f32 * dt;
            self.pos.x = self.pos.x.max(0.0 - 16.0);
            pedal_speed_target = 0.75;
        } else if is_key_down(KeyCode::Right) || is_key_down(KeyCode::D) {
            self.pos.x += Slice::WIDTH / 2.0 * self.speed as f32 * dt;
            self.pos.x = self.pos.x.min(SCREEN_WIDTH - 16.0);
            pedal_speed_target = 1.5;
        }

        if self.pedal_speed_mod < pedal_speed_target {
            self.pedal_speed_mod +=
                (0.5_f64).min((pedal_speed_target - self.pedal_speed_mod) / dt as f64) * dt as f64;
        } else {
            self.pedal_speed_mod -=
                (0.5_f64).min((self.pedal_speed_mod - pedal_speed_target) / dt as f64) * dt as f64;
        }

        self.pedal_speed = self.speed.sqrt() * self.pedal_speed_mod;
        let pedal_speed_visual = self.speed.sqrt() * pedal_speed_target;

        // Iteratively render the trajectory
        let mut cpos = self.pos + 4.0;
        let mut cvel = self.shoot_velocity();
        for i in 0..1000 {
            let dt = 1. / 100.;
            cvel.y += 128.0 * dt;

            if cpos.y > SCREEN_HEIGHT {
                break;
            }

            if i % 10 == 0 {
                draw_hexagon(
                    cpos.x,
                    cpos.y,
                    0.5,
                    0.5,
                    false,
                    Color::new(0.0, 0.0, 0.0, 0.5),
                    Color::new(1.0, 1.0, 1.0, 0.75),
                );
            }
            cpos += cvel * dt;
        }

        let t = get_time();

        // Bike wheels
        draw_texture_ex(
            assets.bike_texture,
            self.pos.x,
            self.pos.y,
            color::WHITE,
            DrawTextureParams {
                source: Some(Rect::new(
                    0.0,
                    (t * 32.0 * pedal_speed_visual % 4.0).floor() as f32 * 32.0,
                    32.0,
                    32.0,
                )),
                ..Default::default()
            },
        );

        // Bike frame
        draw_texture_ex(
            assets.bike_texture,
            self.pos.x,
            self.pos.y,
            color::WHITE,
            DrawTextureParams {
                source: Some(Rect::new(32.0, 32.0, 32.0, 32.0)),
                ..Default::default()
            },
        );

        // Pedals
        draw_texture_ex(
            assets.bike_texture,
            self.pos.x,
            self.pos.y,
            color::WHITE,
            DrawTextureParams {
                source: Some(Rect::new(
                    64.0,
                    (t * 4.0 * pedal_speed_visual % 2.0).floor() as f32 * 32.0,
                    32.0,
                    32.0,
                )),
                ..Default::default()
            },
        );

        // The shadow
        draw_texture_ex(
            assets.bike_texture,
            self.pos.x + 4.0,
            self.pos.y + 29.0,
            Color::new(0.0, 0.0, 0.0, 0.5),
            DrawTextureParams {
                source: Some(Rect::new(32.0, 96.0, 32.0, 32.0)),
                ..Default::default()
            },
        );

        // The character
        draw_texture_ex(
            assets.bike_texture,
            self.pos.x,
            self.pos.y,
            color::WHITE,
            DrawTextureParams {
                source: Some(Rect::new(
                    96.0,
                    (t * 4.0 * pedal_speed_visual % 2.0).floor() as f32 * 32.0,
                    32.0,
                    32.0,
                )),
                ..Default::default()
            },
        );
    }
}

struct World<'a> {
    slices: LinkedList<(f32, Slice)>,
    player: Player,
    projectiles: Vec<Projectile>,
    particles: ParticleWorld<'a>,
    particle_controller_ring: ParticleControllerRing,
    particle_controller_pop_up: ParticleControllerPopUp,
    score: u64,
    score_computed: u64,
}

enum CollisionType {
    Door,
    Window,
}

impl CollisionType {
    fn score(&self) -> u64 {
        match *self {
            CollisionType::Door => 100,
            CollisionType::Window => 200,
        }
    }

    fn particle(&self) -> ParticleType {
        match *self {
            CollisionType::Door => ParticleType::Score100,
            CollisionType::Window => ParticleType::Score200,
        }
    }
}

fn find_collision<'a>(
    projectile: &Projectile,
    slices: impl Iterator<Item = &'a mut (f32, Slice)>,
) -> Option<(&'a mut dyn BoundingBox, Rect, CollisionType)> {
    let angle =
        (-projectile.velocity.angle_between(vec2(1.0, 0.0))).rem_euclid(std::f32::consts::PI * 2.0);

    let check_atttach = |aabb: &dyn BoundingBox, attach_point: Vec2| -> Option<Rect> {
        if let Some(bounds) = aabb.get_bounds() {
            let bounds_real = bounds.offset(attach_point);
            if bounds_real.contains(projectile.pos) && aabb.can_enter_from_dir(angle) {
                return Some(bounds_real);
            }
        }

        None
    };

    for (x, slice) in slices {
        if let Some(ref mut house) = slice.house {
            let house_pos = vec2(0.0, SCREEN_HEIGHT) - Slice::HOUSE_ATTACH_POINT
                + vec2(*x, 0.0)
                + house.attach_offset();

            match house {
                House::Round {
                    door: Attach(door_attach, door),
                    window: Attach(window_attach, window),
                } => {
                    if let Some(bounds_real) = check_atttach(door, house_pos + *door_attach) {
                        return Some((door, bounds_real, CollisionType::Door));
                    }

                    if let Some(bounds_real) = check_atttach(window, house_pos + *window_attach) {
                        return Some((window, bounds_real, CollisionType::Window));
                    }
                }
                House::Wooden {
                    door: Attach(door_attach, door),
                    window_left: Attach(window_left_attach, window_left),
                    window_right: Attach(window_right_attach, window_right),
                } => {
                    if let Some(bounds_real) = check_atttach(door, house_pos + *door_attach) {
                        return Some((door, bounds_real, CollisionType::Door));
                    }

                    if let Some(bounds_real) =
                        check_atttach(window_left, house_pos + *window_left_attach)
                    {
                        return Some((window_left, bounds_real, CollisionType::Window));
                    }

                    if let Some(bounds_real) =
                        check_atttach(window_right, house_pos + *window_right_attach)
                    {
                        return Some((window_right, bounds_real, CollisionType::Window));
                    }
                }
            };
        }
    }

    None
}

impl<'a> World<'a> {
    fn new() -> Self {
        let mut slices = LinkedList::new();

        for i in 0..(SCREEN_WIDTH / 64.0).ceil() as i32 {
            let mut slice = Slice::generate();
            slice.road_border = RoadBorderPosition::None;
            slices.push_back((i as f32 * Slice::WIDTH, slice));
        }

        World {
            slices,
            player: Player::new(),
            projectiles: Vec::new(),
            particles: ParticleWorld::default(),
            particle_controller_ring: ParticleControllerRing,
            particle_controller_pop_up: ParticleControllerPopUp,
            score: 0,
            score_computed: 0,
        }
    }

    fn render(&mut self, assets: &Assets) {
        for (x, _) in self.slices.iter_mut() {
            *x -= self.player.speed as f32 * get_frame_time() * Slice::WIDTH;
        }

        for proj in self.projectiles.iter_mut() {
            proj.pos.x -= self.player.speed as f32 * get_frame_time() * Slice::WIDTH;
        }

        while let Some((x, _)) = self.slices.front_mut() {
            if *x > -Slice::WIDTH * 5.0 {
                break;
            }

            self.slices.pop_front();
        }

        while let Some((x, _)) = self.slices.back() {
            if *x < SCREEN_WIDTH + Slice::WIDTH {
                self.slices.push_back((x + Slice::WIDTH, Slice::generate()));
                continue;
            }

            break;
        }

        for (x, slice) in self.slices.iter_mut() {
            slice.render(vec2(*x, SCREEN_HEIGHT), assets);
        }

        if is_key_pressed(KeyCode::Space) && self.player.lives > 0 {
            play_sound_once(assets.shoot_sound);

            self.projectiles.push(Projectile::new(
                self.player.pos,
                self.player.shoot_velocity(),
            ));

            self.particles.spawn(
                ConfigRingBasic {
                    particle_type: ParticleType::Explosion,
                    pos: self.player.pos,
                    fade_start: 0.5,
                    lifetime: 0.5,
                    count: 10,
                    radius: 5.0..7.0,
                    velocity: 10.0,
                    scale: 1.0,
                    color: color::WHITE,
                    angle_cone: Some(0.0..std::f32::consts::FRAC_PI_2),
                },
                &self.particle_controller_ring,
            );
        }

        let dt = get_frame_time();
        self.particles.render(dt, assets);

        self.projectiles.retain_mut(|p| {
            if let Some((hit, bounds, collision_type)) = find_collision(p, self.slices.iter_mut()) {
                let rad = bounds.size().length() / 2.0;
                self.particles.spawn(
                    ConfigRingBasic {
                        particle_type: ParticleType::Spark,
                        pos: bounds.center(),
                        lifetime: 0.75,
                        count: 20,
                        radius: rad * 0.75..rad * 1.25,
                        velocity: 5.0,
                        fade_start: 0.6,
                        scale: 0.5,
                        color: color::WHITE,
                        angle_cone: None,
                    },
                    &self.particle_controller_ring,
                );

                let pos = Vec2::new(bounds.x + bounds.w / 2.0, bounds.y - 4.0);

                hit.replace_with_closed();

                self.score += collision_type.score();

                self.particles.spawn(
                    ConfigPopUpBasic {
                        particle_type: collision_type.particle(),
                        pos,
                        lifetime: 0.75,
                        velocity: vec2(0.0, -10.0),
                        accel: vec2(0.0, 20.0),
                        fade_start: 0.6,
                        scale: 0.5,
                        color: color::WHITE,
                    },
                    &self.particle_controller_pop_up,
                );

                play_sound_once(assets.hit_sound);

                return false;
            }

            p.render(assets);

            p.particle_timer += dt;

            if p.particle_timer > 0.05 {
                let mut color = color::hsl_to_rgb(f32::gen_range(0.0, 1.0), 1.0, 0.5);
                color.a = 0.5;
                self.particles.spawn(
                    ConfigRingBasic {
                        particle_type: ParticleType::Heart,
                        pos: p.pos,
                        lifetime: 0.75,
                        fade_start: 0.3,
                        count: 1,
                        radius: 0.0..1.0,
                        velocity: 2.0,
                        scale: 1.0,
                        color,
                        angle_cone: None,
                    },
                    &self.particle_controller_ring,
                );
                p.particle_timer = 0.0;
            }

            if p.pos.y < SCREEN_HEIGHT + 20.0 {
                true
            } else {
                self.score = self.score.saturating_sub(50);

                self.particles.spawn(
                    ConfigRingBasic {
                        particle_type: ParticleType::Explosion,
                        pos: vec2(p.pos.x, SCREEN_HEIGHT + 5.0),
                        lifetime: 1.0,
                        fade_start: 0.7,
                        count: 20,
                        radius: 0.0..1.0,
                        velocity: 25.0,
                        scale: 0.5,
                        color: color::WHITE,
                        angle_cone: Some(
                            std::f32::consts::FRAC_PI_2 - 0.5..std::f32::consts::FRAC_PI_2 + 0.5,
                        ),
                    },
                    &self.particle_controller_ring,
                );

                play_sound_once(assets.miss_sound);

                false
            }
        });

        if self.player.lives > 0 {
            for (x, slice) in self.slices.iter_mut() {
                let slice_pos = vec2(0.0, SCREEN_HEIGHT) + vec2(*x, 0.0);

                match slice.road_border {
                    RoadBorderPosition::None => {}
                    RoadBorderPosition::Top(_) => {
                        if Rect::new(
                            slice_pos.x + 12.0 - 30.0,
                            slice_pos.y - 54.0 - 29.0,
                            12.0,
                            25.0,
                        )
                        .contains(self.player.pos)
                        {
                            self.player.lives -= 1;
                            play_sound_once(assets.lose_life_sound);
                            slice.road_border = RoadBorderPosition::None;
                        }
                    }
                    RoadBorderPosition::Bottom(_) => {
                        if Rect::new(
                            slice_pos.x + 12.0 - 30.0,
                            slice_pos.y - 25.0 - 29.0,
                            12.0,
                            25.0,
                        )
                        .contains(self.player.pos)
                        {
                            self.player.lives -= 1;
                            play_sound_once(assets.lose_life_sound);
                            slice.road_border = RoadBorderPosition::None;
                        }
                    }
                };
            }

            self.player.render(assets);
        }

        for (x, slice) in self.slices.iter_mut() {
            slice.render_front(vec2(*x, SCREEN_HEIGHT), assets);
        }

        if self.player.lives > 0 {
            self.score_computed = self.score + (self.player.speed * 1000.0).floor() as u64;
        }

        // Let's say the player is roughly 1.70m
        let px_to_m: f64 = 15.0 / Slice::WIDTH as f64;
        let speed_text = format!(
            "Speed: {:.2}km/h",
            self.player.speed * Slice::WIDTH as f64 * px_to_m
        );
        draw_string(&speed_text, 2.0, 20.4, color::BLACK, assets.font);
        draw_string(&speed_text, 2.0, 20.0, color::PURPLE, assets.font);

        let score_text = format!("Score: {:09}", self.score_computed);
        draw_string(&score_text, 2.0, 10.4, color::BLACK, assets.font);
        draw_string(&score_text, 2.0, 10.0, color::WHITE, assets.font);

        let lives_text = "Lives:";
        draw_string(
            lives_text,
            SCREEN_WIDTH - 60.0,
            10.4,
            color::BLACK,
            assets.font,
        );
        draw_string(
            lives_text,
            SCREEN_WIDTH - 60.0,
            10.0,
            color::WHITE,
            assets.font,
        );

        for i in 0..self.player.lives {
            draw_texture_ex(
                assets.blobcat_texture,
                SCREEN_WIDTH - 60.0 + 10.0 * i as f32,
                12.0,
                color::WHITE,
                DrawTextureParams {
                    dest_size: Some(vec2(16.0, 16.0)),
                    ..Default::default()
                },
            );
        }
    }
}

enum ProjectileType {
    Cat,
    Fox,
    Corgi,
}

impl ProjectileType {
    fn generate() -> Self {
        match u32::gen_range(0, 100) {
            ..=70 => ProjectileType::Cat,
            71..=98 => ProjectileType::Fox,
            _ => ProjectileType::Corgi,
        }
    }

    fn sprite_rect(&self) -> Rect {
        match *self {
            ProjectileType::Cat => Rect::new(0.0, 0.0, 16.0, 16.0),
            ProjectileType::Fox => Rect::new(17.0, 0.0, 16.0, 16.0),
            ProjectileType::Corgi => Rect::new(18.0, 0.0, 16.0, 16.0),
        }
    }
}

pub struct Projectile {
    projectile_type: ProjectileType,
    pos: Vec2,
    velocity: Vec2,
    rot: f32,
    rot_velocity: f32,
    particle_timer: f32,
}

impl Projectile {
    fn new(pos: Vec2, velocity: Vec2) -> Self {
        Projectile {
            projectile_type: ProjectileType::generate(),
            pos,
            velocity,
            rot: f32::gen_range(0.0, std::f32::consts::PI * 2.0),
            rot_velocity: f32::gen_range(-std::f32::consts::PI, std::f32::consts::PI),
            particle_timer: 0.01,
        }
    }

    fn render(&mut self, assets: &Assets) {
        let dt = get_frame_time();

        self.velocity.y += 128.0 * dt;
        self.pos += self.velocity * dt;
        self.rot += self.rot_velocity * dt;

        let sprite_rect = self.projectile_type.sprite_rect();

        draw_texture_ex(
            assets.projectiles_texture,
            self.pos.x,
            self.pos.y,
            color::WHITE,
            DrawTextureParams {
                source: Some(sprite_rect),
                rotation: self.rot,
                dest_size: Some(sprite_rect.size() / 2.0),
                ..Default::default()
            },
        );
    }
}

pub struct GameWorld<'a> {
    sky: Sky,
    world: World<'a>,
}

impl Default for GameWorld<'_> {
    fn default() -> Self {
        GameWorld {
            sky: Sky::default(),
            world: World::new(),
        }
    }
}

impl<'a> GameWorld<'a> {
    pub async fn frame(&mut self, assets: &Assets) {
        self.sky.render(assets);

        self.world.render(assets);
    }
}
