use crate::assets::{draw_string, Assets};
use crate::config::{SCREEN_HEIGHT, SCREEN_WIDTH};
use crate::sky::Sky;
use macroquad::color;
use macroquad::input::{is_key_pressed, KeyCode};
use macroquad::math::{vec2, Rect, Vec2};
use macroquad::prelude::draw_texture_ex;
use macroquad::texture::{draw_texture, DrawTextureParams};
use macroquad::time::get_frame_time;

struct BlobCat {
    pos: Vec2,
    velocity: Vec2,
    rot: f32,
}

impl Default for BlobCat {
    fn default() -> Self {
        BlobCat {
            pos: vec2(25.0, 100.0),
            velocity: vec2(25.0, 25.0),
            rot: 0.0,
        }
    }
}

impl BlobCat {
    fn render(&mut self, assets: &Assets) {
        if self.pos.x < 0.0 {
            self.velocity.x = self.velocity.x.abs();
        }

        if self.pos.x > SCREEN_WIDTH - 32.0 {
            self.velocity.x = -self.velocity.x.abs();
        }

        if self.pos.y < 0.0 {
            self.velocity.y = self.velocity.y.abs();
        }

        if self.pos.y > SCREEN_HEIGHT - 32.0 {
            self.velocity.y = -self.velocity.y.abs();
        }

        let dt = get_frame_time();
        self.pos += self.velocity * dt;
        self.rot += dt;

        draw_texture_ex(
            assets.blobcat_texture,
            self.pos.x,
            self.pos.y,
            color::WHITE,
            DrawTextureParams {
                rotation: self.rot,
                ..Default::default()
            },
        );
    }
}

#[derive(Default)]
pub struct GameMenu {
    sky: Sky,
    blobcat: BlobCat,
}

impl GameMenu {
    pub async fn frame(&mut self, assets: &Assets) -> bool {
        self.sky.render(assets);

        self.blobcat.render(assets);

        draw_texture(
            assets.menu_texture,
            SCREEN_WIDTH / 2.0 - 64.0,
            SCREEN_HEIGHT / 2.0 - 95.0,
            color::BLACK,
        );
        draw_texture(
            assets.menu_texture,
            SCREEN_WIDTH / 2.0 - 64.0,
            SCREEN_HEIGHT / 2.0 - 96.0,
            color::WHITE,
        );

        draw_texture_ex(
            assets.controls_texture,
            SCREEN_WIDTH / 2.0 - 96.0,
            SCREEN_HEIGHT / 2.0 - 5.0,
            color::WHITE,
            DrawTextureParams {
                source: Some(Rect::new(0.0, 0.0, 106.0, 32.0)),
                ..Default::default()
            },
        );

        let move_text = "Movement";
        let move_x = SCREEN_WIDTH / 2.0 - 60.0;
        let move_y = SCREEN_HEIGHT / 2.0 + 38.0;
        draw_string(move_text, move_x, move_y + 0.4, color::BLACK, assets.font);
        draw_string(move_text, move_x, move_y, color::WHITE, assets.font);

        draw_texture_ex(
            assets.controls_texture,
            SCREEN_WIDTH / 2.0 + 24.0,
            SCREEN_HEIGHT / 2.0 + 13.0,
            color::WHITE,
            DrawTextureParams {
                source: Some(Rect::new(0.0, 32.0, 48.0, 16.0)),
                ..Default::default()
            },
        );

        let shoot_text = "Shoot";
        let shoot_x = SCREEN_WIDTH / 2.0 + 35.0;
        let shoot_y = SCREEN_HEIGHT / 2.0 + 38.0;
        draw_string(
            shoot_text,
            shoot_x,
            shoot_y + 0.4,
            color::BLACK,
            assets.font,
        );
        draw_string(shoot_text, shoot_x, shoot_y, color::WHITE, assets.font);

        let play_text = "Press [Space] to play...";
        let play_x = SCREEN_WIDTH / 2.0 - 40.0;
        let play_y = SCREEN_HEIGHT / 2.0 + 60.0;
        draw_string(play_text, play_x, play_y + 0.4, color::BLACK, assets.font);
        draw_string(play_text, play_x, play_y, color::WHITE, assets.font);

        is_key_pressed(KeyCode::Space)
    }
}
