use crate::assets::Assets;
use macroquad::color;
use macroquad::color::Color;
use macroquad::logging::debug;
use macroquad::math::{vec2, Rect, Vec2};
use macroquad::prelude::DrawTextureParams;
use macroquad::rand::RandomRange;
use macroquad::texture::draw_texture_ex;
use std::ops::Range;

#[derive(Default)]
pub struct ParticleWorld<'a> {
    particles: Vec<Particle<'a>>,
}

impl ParticleWorld<'_> {
    pub fn spawn<C>(
        &mut self,
        config: C,
        controller: &'_ (impl ParticleController + ParticleSpawner<Config = C>),
    ) {
        controller.spawn(config, |c| self.particles.push(c));
    }

    pub fn render(&mut self, dt: f32, assets: &Assets) {
        self.particles.retain_mut(|particle| {
            !matches!(
                particle.controller.render(dt, particle, assets),
                ParticleLifetime::Destroy
            )
        })
    }
}

#[derive(Copy, Clone)]
pub enum ParticleType {
    Spark,
    Heart,
    Explosion,
    Score100,
    Score200,
    Fail,
}

const SPARK_FRAMES: &[Rect] = &[
    Rect {
        x: 0.0,
        y: 0.0,
        w: 8.0,
        h: 8.0,
    },
    Rect {
        x: 8.0,
        y: 0.0,
        w: 8.0,
        h: 8.0,
    },
    Rect {
        x: 16.0,
        y: 0.0,
        w: 8.0,
        h: 8.0,
    },
    Rect {
        x: 24.0,
        y: 0.0,
        w: 8.0,
        h: 8.0,
    },
];

const HEART_FRAMES: &[Rect] = &[
    Rect {
        x: 0.0,
        y: 8.0,
        w: 8.0,
        h: 8.0,
    },
    Rect {
        x: 8.0,
        y: 8.0,
        w: 8.0,
        h: 8.0,
    },
    Rect {
        x: 16.0,
        y: 8.0,
        w: 8.0,
        h: 8.0,
    },
    Rect {
        x: 24.0,
        y: 8.0,
        w: 8.0,
        h: 8.0,
    },
];

const EXPLOSION_FRAMES: &[Rect] = &[
    Rect {
        x: 0.0,
        y: 16.0,
        w: 8.0,
        h: 8.0,
    },
    Rect {
        x: 8.0,
        y: 16.0,
        w: 8.0,
        h: 8.0,
    },
    Rect {
        x: 16.0,
        y: 16.0,
        w: 8.0,
        h: 8.0,
    },
    Rect {
        x: 24.0,
        y: 16.0,
        w: 8.0,
        h: 8.0,
    },
];

const SCORE100_FRAMES: &[Rect] = &[Rect {
    x: 1.0,
    y: 25.0,
    w: 27.0,
    h: 15.0,
}];

const SCORE200_FRAMES: &[Rect] = &[Rect {
    x: 33.0,
    y: 25.0,
    w: 27.0,
    h: 15.0,
}];

const FAIL_FRAMES: &[Rect] = &[Rect {
    x: 32.0,
    y: 9.0,
    w: 14.0,
    h: 14.0,
}];

impl ParticleType {
    fn frames(&self) -> &[Rect] {
        match *self {
            ParticleType::Spark => SPARK_FRAMES,
            ParticleType::Heart => HEART_FRAMES,
            ParticleType::Explosion => EXPLOSION_FRAMES,
            ParticleType::Score100 => SCORE100_FRAMES,
            ParticleType::Score200 => SCORE200_FRAMES,
            ParticleType::Fail => FAIL_FRAMES,
        }
    }
}

#[derive(Copy, Clone, Default)]
pub enum AnimationMode {
    #[default]
    TieToLifetime,
    Repeat(f32),
}

#[derive(Default)]
pub struct ParticleState {
    pos: Vec2,
    velocity: Vec2,
    accel: Vec2,
    rot: f32,
    rot_velocity: f32,
    lifetime: f32,
    max_lifetime: f32,
    color: Color,
    animation_mode: AnimationMode,
    scale: f32,
    fade_start: f32,
}

pub struct Particle<'a> {
    particle_type: ParticleType,
    state: ParticleState,
    controller: &'a dyn ParticleController,
}

pub enum ParticleLifetime {
    Keep,
    Destroy,
}

pub trait ParticleSpawner {
    type Config;

    fn spawn<'a>(&'a self, config: Self::Config, consumer: impl FnMut(Particle<'a>));
}

pub trait ParticleController {
    fn render(&self, dt: f32, particle: &mut Particle<'_>, assets: &Assets) -> ParticleLifetime;
}

// Implementations!

fn tick(state: &mut ParticleState, dt: f32) {
    state.lifetime += dt;
    state.velocity += state.accel * dt;
    state.pos += state.velocity * dt;
    state.rot += state.rot_velocity * dt;
    state.color.a = ((state.fade_start - state.lifetime) / (state.max_lifetime - state.fade_start))
        .max(0.0)
        .min(1.0)
}

fn render_quad(particle_type: ParticleType, state: &ParticleState, assets: &Assets) {
    let sprite_rect = particle_type.frames();

    let frame = match state.animation_mode {
        AnimationMode::TieToLifetime if state.max_lifetime <= 0.0 => sprite_rect[0],
        AnimationMode::TieToLifetime => {
            sprite_rect[((state.lifetime / state.max_lifetime) * sprite_rect.len() as f32).floor()
                as usize
                % sprite_rect.len()]
        }
        AnimationMode::Repeat(loop_len) => {
            sprite_rect[((state.lifetime / loop_len % 1.0) * sprite_rect.len() as f32).floor()
                as usize
                % sprite_rect.len()]
        }
    };

    draw_texture_ex(
        assets.particles_texture,
        state.pos.x,
        state.pos.y,
        state.color,
        DrawTextureParams {
            dest_size: Some(frame.size() * state.scale),
            rotation: state.rot,
            source: Some(frame),
            ..Default::default()
        },
    );
}

// Ring

#[derive(Clone)]
pub struct ConfigRingBasic {
    pub particle_type: ParticleType,
    pub pos: Vec2,
    pub lifetime: f32,
    pub count: usize,
    pub radius: Range<f32>,
    pub velocity: f32,
    pub scale: f32,
    pub color: Color,
    pub fade_start: f32,
    pub angle_cone: Option<Range<f32>>,
}

pub struct ParticleControllerRing;

impl ParticleSpawner for ParticleControllerRing {
    type Config = ConfigRingBasic;

    fn spawn<'a>(&'a self, config: Self::Config, mut consumer: impl FnMut(Particle<'a>)) {
        for _ in 0..config.count {
            let angle = if let Some(ref cone) = config.angle_cone {
                f32::gen_range(cone.start, cone.end)
            } else {
                f32::gen_range(0.0, 2.0 * std::f32::consts::PI)
            };
            let r = f32::gen_range(config.radius.start, config.radius.end);

            let cos_t = angle.cos();
            let sin_t = -angle.sin();

            let x = config.pos.x + cos_t * r;
            let y = config.pos.y + sin_t * r;

            let vel = vec2(cos_t, sin_t) * config.velocity;

            consumer(Particle {
                particle_type: config.particle_type,
                state: ParticleState {
                    pos: vec2(x, y),
                    velocity: vel,
                    rot: f32::gen_range(0.0, 2.0 * std::f32::consts::PI),
                    color: config.color,
                    max_lifetime: config.lifetime,
                    scale: config.scale,
                    fade_start: config.fade_start,
                    ..Default::default()
                },
                controller: self,
            });
        }
    }
}

impl ParticleController for ParticleControllerRing {
    fn render(&self, dt: f32, particle: &mut Particle<'_>, assets: &Assets) -> ParticleLifetime {
        tick(&mut particle.state, dt);

        if particle.state.lifetime > particle.state.max_lifetime {
            ParticleLifetime::Destroy
        } else {
            render_quad(particle.particle_type, &particle.state, assets);
            ParticleLifetime::Keep
        }
    }
}

// PopUp

#[derive(Clone)]
pub struct ConfigPopUpBasic {
    pub particle_type: ParticleType,
    pub pos: Vec2,
    pub lifetime: f32,
    pub velocity: Vec2,
    pub scale: f32,
    pub color: Color,
    pub fade_start: f32,
    pub accel: Vec2,
}

pub struct ParticleControllerPopUp;

impl ParticleSpawner for ParticleControllerPopUp {
    type Config = ConfigPopUpBasic;

    fn spawn<'a>(&'a self, config: Self::Config, mut consumer: impl FnMut(Particle<'a>)) {
        consumer(Particle {
            particle_type: config.particle_type,
            state: ParticleState {
                pos: config.pos,
                velocity: config.velocity,
                color: config.color,
                max_lifetime: config.lifetime,
                scale: config.scale,
                fade_start: config.fade_start,
                accel: config.accel,
                ..Default::default()
            },
            controller: self,
        });
    }
}

impl ParticleController for ParticleControllerPopUp {
    fn render(&self, dt: f32, particle: &mut Particle<'_>, assets: &Assets) -> ParticleLifetime {
        tick(&mut particle.state, dt);

        if particle.state.lifetime > particle.state.max_lifetime {
            ParticleLifetime::Destroy
        } else {
            render_quad(particle.particle_type, &particle.state, assets);
            ParticleLifetime::Keep
        }
    }
}
